{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
{*******************************************************}
{                                                       }
{                      Tachyon Unit                     }
{    Vector Raster Geographic Information Synthesis     }
{                     VOICE  ..  Tracer                 }
{                     GRIP ICE .. Tongs                 }
{                Digital Terrain Mapping                }
{               Image Locatable Holographics            }
{                          SOS MAP                      }
{  Surreal Object Synthesis Multimedia Analysis Product }
{                   Fractal3D  Life MOW                 }
{       Copyright (c) 1995,2006  Ivan Lee Herring       }
{                                                       }
{*******************************************************}
unit buzAlber;
{*************************************************************************
NAME                     ALBERS CONICAL EQUAL AREA
PURPOSE:	Transforms input longitude and latitude to Easting and Northing
  for the Albers Conical Equal Area projection.  The longitude
  and latitude must be in radians.  The Easting and Northing
  values will be returned in meters.
PROGRAMMER              DATE
----------              ----
T. Mittan,       	Feb, 1992
ALGORITHM REFERENCES
1.  Snyder, John P., "Map Projections--A Working Manual", U.S. Geological
    Survey Professional Paper 1395 (Supersedes USGS Bulletin 1532), United
    State Government Printing Office, Washington D.C., 1987.
2.  Snyder, John P. and Voxland, Philip M., "An Album of Map Projections",
    U.S. Geological Survey Professional Paper 1453 , United State Government
    Printing Office, Washington D.C., 1989.
***********************************************************************
#include "gctpc/cproj.h"
#include "gctpc/report.h"}

interface

uses Math;

//* Variables common to ALBERS CONICAL EQUAL AREA
var
  r_major, //* major axis
    r_minor, //* minor axis
    c, //* constant c
    e3, //* eccentricity
    rh, //* heigth above elipsoid
    ns0, //* ratio between meridians
    lon_center, //* center longitude
    false_easting, //* x offset in meters
    false_northing: Double; //* y offset in meters

function
  alberforint(r_maj, r_min, lat1, lat2, lon0, lat0,
  false_east, false_north: double): Integer;

function
  alberfor(lon, lat: double; var x, y: double): Integer;

function
  alberinvint(r_maj, r_min, lat1, lat2, lon0,
  lat0, false_east, false_north: double): Integer;

function
  alberinv(x, y: double; var lon, lat: double): Integer;

implementation

uses buzGlobal;

function
  alberforint(r_maj, r_min, lat1, lat2, lon0, lat0,
  false_east, false_north: double): Integer;
{ Initialize the Albers projection
long alberforint(
double r_maj,/* major axis
double r_min,/* minor axis
double lat1,/* first standard parallel
double lat2,/* second standard parallel
double lon0,/* center longitude
  double lat0,/* center lattitude
 double false_east,/* x offset in meters
double false_north)/* y offset in meters}
{double msfnz(double eccent, double sinphi, double cosphi);
/* function to compute small m		*/
double qsfnz(double eccent, double sinphi, double cosphi);
 /* function to compute small q		*/}
var
  sin_po, cos_po, //* sin and cos values
    con, //* temporary variable
    es, temp, //* eccentricity squared and temp var
    ms1, //* small m 1
    ms2, //* small m 2
    qs0, //* small q 0
    qs1, //* small q 1
    qs2: double; //* small q 2
begin
  false_easting := false_east;
  false_northing := false_north;
  lon_center := lon0;
  if (abs(lat1 + lat2) < EPSLN) then
  begin
{    PError('Equal latitudes for St. Parallels on opposite sides of equator',
      'alber-forinit'); }
  DoProjError(31,31,0,0,0);
    result := (31);
    exit;
  end;
  r_major := r_maj;
  r_minor := r_min;
  temp := r_minor / r_major;
  es := 1.0 - SQR(temp);
  e3 := sqrt(es);
  sin_po := Sin(lat1);
  cos_po := Cos(lat1);
  {sincos(lat1, sin_po, cos_po);}
  con := sin_po;

  ms1 := msfnz(e3, sin_po, cos_po);
  qs1 := qsfnz(e3, sin_po, cos_po);
  sin_po := Sin(lat2);
  cos_po := Cos(lat2);
  {sincos(lat2, sin_po, cos_po);}

  ms2 := msfnz(e3, sin_po, cos_po);
  qs2 := qsfnz(e3, sin_po, cos_po);
  sin_po := Sin(lat0);
  cos_po := Cos(lat0);
  {sincos(lat0, sin_po, cos_po);}

  qs0 := qsfnz(e3, sin_po, cos_po);

  if (abs(lat1 - lat2) > EPSLN) then
    ns0 := (ms1 * ms1 - ms2 * ms2) / (qs2 - qs1)
  else
    ns0 := con;
  c := ms1 * ms1 + ns0 * qs1;
  rh := r_major * sqrt(c - ns0 * qs0) / ns0;

//* Report parameters to the user  -----------------------------*/
  ptitle('ALBERS CONICAL EQUAL-AREA');
  radius2(r_major, r_minor);
  stanparl(lat1, lat2);
  cenlonmer(lon_center);
  origin(lat0);
  offsetp(false_easting, false_northing);

  result := (OK);
end;


function
  alberfor(lon, lat: double; var x, y: double): Integer;
{ Albers Conical Equal Area forward equations--mapping lat,long to x,y
long alberfor(
double lon,/* (I) Longitude
double lat,/* (I) Latitude
double *x,/* (O) X projection coordinate
double *y)/* (O) Y projection coordinate}
var
  sin_phi, cos_phi, //* sine and cos values		*/
    qs, //* small q			*/
    theta, //* angle			*/
    rh1: double; //* height above ellipsoid	*/
{double qsfnz(double eccent, double sinphi, double cosphi);
 /* function to calculate q	*/
double adjust_lon(double x);
 /* Function to adjust longitude to -180 - 180   */}
begin
  sin_phi := Sin(lat);
  cos_phi := Cos(lat);
  {sincos(lat, sin_phi, cos_phi);}
  qs := qsfnz(e3, sin_phi, cos_phi);
  rh1 := r_major * sqrt(c - ns0 * qs) / ns0;
  theta := ns0 * AdjustLon(lon - lon_center);
  x := rh1 * sin(theta) + false_easting;
  y := rh - rh1 * cos(theta) + false_northing;
  result := (OK);
end;


function
  alberinvint(r_maj, r_min, lat1, lat2, lon0,
  lat0, false_east, false_north: double): Integer;
{ Initialize the Albers projection  -------------------------------*/
long alberinvint(
double r_maj,/* major axis
double r_min,/* minor axis
double lat1, /* first standard parallel
 double lat2,/* second standard parallel
 double lon0,/* center longitude
 double lat0,/* center lattitude
 double false_east,/* x offset in meters
double false_north)/* y offset in meters}
{double msfnz(double eccent, double sinphi, double cosphi);
 /* function to compute small m		*/
double qsfnz(double eccent, double sinphi, double cosphi);
 /* function to compute small q		*/}
var
  sin_po, cos_po, //* sine and cos values
    con, //* temporary variable
    temp, //* temporary variable
    ms1, //* small m 1
    ms2, //* small m 2
    qs0, //* small q 0
    qs1, //* small q 1
    es, qs2: double; //* small q 2
begin
  false_easting := false_east;
  false_northing := false_north;
  lon_center := lon0;
  if (abs(lat1 + lat2) < EPSLN) then
  begin
{    PError('Equal latitudes for Standard Parallels on opposite sides of equator'
      , 'alber-invinit');}
  DoProjError(32,31,0,0,0);
    result := (31);
    exit;
  end;
  r_major := r_maj;
  r_minor := r_min;
  temp := r_minor / r_major;
  es := 1.0 - SQR(temp);
  e3 := sqrt(es);
  sin_po := Sin(lat1);
  cos_po := Cos(lat1);
  {sincos(lat1, sin_po, cos_po);}
  con := sin_po;

  ms1 := msfnz(e3, sin_po, cos_po);
  qs1 := qsfnz(e3, sin_po, cos_po);
  sin_po := Sin(lat2);
  cos_po := Cos(lat2);
  {sincos(lat2, sin_po, cos_po);}

  ms2 := msfnz(e3, sin_po, cos_po);
  qs2 := qsfnz(e3, sin_po, cos_po);
  sin_po := Sin(lat0);
  cos_po := Cos(lat0);
  {sincos(lat0, sin_po, cos_po);}

  qs0 := qsfnz(e3, sin_po, cos_po);

  if (abs(lat1 - lat2) > EPSLN) then
    ns0 := (ms1 * ms1 - ms2 * ms2) / (qs2 - qs1)
  else ns0 := con;
  c := ms1 * ms1 + ns0 * qs1;
  rh := r_major * sqrt(c - ns0 * qs0) / ns0;

//* Report parameters to the user  -----------------------------*/
  ptitle('ALBERS CONICAL EQUAL-AREA');
  radius2(r_major, r_minor);
  stanparl(lat1, lat2);
  cenlonmer(lon_center);
  origin(lat0);
  offsetp(false_easting, false_northing);

  result := OK;
end;


function
  alberinv(x, y: double; var lon, lat: double): Integer;
{Albers Conical Equal Area inverse equations--mapping x,y to lat/long
long alberinv(
double x,/* (O) X projection coordinate
double y,/* (O) Y projection coordinate
double *lon,/* (I) Longitude
double *lat)/* (I) Latitude}
{double qsfnz(double eccent, double sinphi, double cosphi);
/* function to calculate q	*/
double adjust_lon(double x);    /* Function to adjust longitude to -180 - 180   */
double phi1z (double eccent, double qs, long int *flag);
/* function to calculate LAT phi	*/}
var
  rh1, //* height above ellipsoid
  {es,} qs, //* function q
    con, //* temporary sign value
    theta: double; //* angle
  flag: Integer; //* error flag;
begin
  result := (OK);
  flag := 0;
  x := x - false_easting;
  y := rh - y + false_northing; ;
  if (ns0 >= 0) then
  begin
    rh1 := sqrt(x * x + y * y);
    con := 1.0;
  end
  else
  begin
    rh1 := -sqrt(x * x + y * y);
    con := -1.0;
  end;
  theta := 0.0;
  if (rh1 <> 0.0) then theta := arctan((con * x) / (con * y));
  con := rh1 * ns0 / r_major;
  qs := (c - con * con) / ns0;
  if (e3 >= 1E-10) then
  begin {really qs ??? es}
    con := 1 - 0.5 * (1.0 - qs) * ln((1.0 - e3) / 1.0 + e3) / e3;
    if (abs(abs(con) - abs(qs)) > 0.0000000001) then
    begin
      lat := phi1z(e3, qs, flag);
      if (flag <> 0) then result := (flag);
    end
    else
    begin
      if (qs >= 0) then
        lat := 0.5 * PI
      else
        lat := -0.5 * PI;
    end;
  end
  else
  begin
    lat := phi1z(e3, qs, flag);
    if (flag <> 0) then result := (flag);
  end;
  lon := AdjustLon(theta / ns0 + lon_center);
{result:=(OK); moved to top}
end;

end.
